#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "types.h"
#include "nlllib.h"
#include "memory.h"
#include "string.h"
#include "array.h"
#include "area.h"
#include "key.h"
#include "framebuf.h"
#include "value.h"

static int value_num = 0;
static int pool_num = 0;

static value_t pool = NULL;

static value_t pool_alloc(void)
{
  value_t value;

  if (!pool)
    return NULL;

  value = pool;
  pool = pool->next;
  pool_num--;

  return value;
}

static int pool_free(value_t value)
{
  value->next = pool;
  pool = value;
  pool_num++;
  return 0;
}

#ifndef NLL_MEMORY_STATIC
static int _init(void)
{
  return 0;
}

static int _done(void)
{
  while (pool)
    memory_free(pool_alloc());
  return 0;
}

static int _check(void)
{
  if (pool)
    return NLL_ERRCODE_VALUE_NOT_EMPTY;
  return 0;
}

static value_t _alloc(void)
{
  value_t value;
  value = pool_alloc();
  if (!value)
    value = memory_alloc(sizeof(*value));
  return value;
}

static int _free(value_t value)
{
  if ((VALUE_MAXNUM < 0) || (pool_num < VALUE_MAXNUM))
    return pool_free(value);
  memory_free(value);
  return 0;
}
#else
static struct value values[VALUE_MAXNUM];

static int _init(void)
{
  value_t value;
  int i;

  memset(values, 0, sizeof(values));

  for (i = 0; i < VALUE_MAXNUM; i++) {
    value = &values[i];
    value->next = pool;
    pool = value;
    pool_num++;
  }

  return 0;
}

static int _done(void)
{
  return 0;
}

static int _check(void)
{
  value_t value;
  int n = 0;

  for (value = pool; value && (n < VALUE_MAXNUM); value = value->next)
    n++;

  if (value || (n != VALUE_MAXNUM) || (n != pool_num))
    return NLL_ERRCODE_VALUE_NOT_EMPTY;

  return 0;
}

static value_t _alloc(void)
{
  return pool_alloc();
}

static int _free(value_t value)
{
  return pool_free(value);
}
#endif

int value_check(void)
{
  if (value_num)
    return NLL_ERRCODE_VALUE_NOT_EMPTY;

  return _check();
}

int value_alloc(value_t *valuep)
{
  value_t value;

  if ((value = _alloc()) == NULL)
    return NLL_ERRCODE_VALUE_BUFFER_OVER;
  value_num++;

  memset(value, 0, sizeof(*value));

  value->refcount++;

  *valuep = value;

  return 0;
}

int value_free(value_t value)
{
  value_t next;
  int r;

  if (value->refcount > 1) {
    value->refcount--;
    return 0;
  }

  for (; value; value = next) {
    next = value->next;
    value->next = NULL;

    if (--(value->refcount) == 0) {
      if (value->flags & VALUE_FLAG_CONST)
	value->flags &= ~VALUE_FLAG_CONST;
      if ((r = value_clear(value)) < 0)
	return r;

      value_num--;
      if ((r = _free(value)) < 0)
	return r;
    }
  }

  return 0;
}

int value_init(void)
{
  value_num = 0;
  pool_num = 0;
  pool = NULL;
  return _init();
}

int value_done(void)
{
  return _done();
}

int value_clear(value_t value)
{
  int r;

  if (value->flags & VALUE_FLAG_CONST)
    return 0;

  switch (value->type) {
  case VALUE_TYPE_ARRAY:
    if ((r = array_free(value->val.array.array, value->val.array.number)) < 0)
      return r;
    break;

  case VALUE_TYPE_STRING:
    if ((r = string_del(value->val.string)) < 0)
      return r;
    break;

  case VALUE_TYPE_VALUE:
    if ((r = value_free(value->val.value)) < 0)
      return r;
    break;

  case VALUE_TYPE_AREA:
    if ((r = area_free(value->val.area)) < 0)
      return r;
    break;

  case VALUE_TYPE_POINTER:
    if (value->val.pointer.point) {
      if ((r = value_free(value->val.pointer.point)) < 0)
	return r;
    }
    break;

  default:
    break;
  }

  value->type = VALUE_TYPE_UNDEFINED;

  return 0;
}

int value_is_nzero(value_t value)
{
  int r, length;
  value_type_t type;
  integer_t integer;
  char *string;
#ifdef NLL_FLOATING_POINT
  double f;
#endif
  area_t area;
  value_t point;

  type = value_get_type(value);

  switch (type) {
  case VALUE_TYPE_INTEGER:
    if ((r = value_get_integer(value, &integer)) < 0)
      return r;
    return integer ? 1 : 0;

  case VALUE_TYPE_STRING:
    if ((r = value_get_string(value, &string, &length)) < 0)
      return r;
    return length ? 1 : 0;

#ifdef NLL_FLOATING_POINT
  case VALUE_TYPE_FLOAT:
    if ((r = value_get_float(value, &f)) < 0)
      return r;
    return (f != 0.0) ? 1 : 0;
#endif

  case VALUE_TYPE_AREA:
    if ((r = value_get_area(value, &area)) < 0)
      return r;
    return area ? 1 : 0;

  case VALUE_TYPE_POINTER:
    if ((r = value_get_pointer(value, &point, NULL)) < 0)
      return r;
    return point ? 1 : 0;

  default:
    break;
  }

  return NLL_ERRCODE_VALUE_INVALID_TYPE;
}

value_t value_entity(value_t value)
{
  while (value->type == VALUE_TYPE_VALUE)
    value = value->val.value;

  return value;
}

int value_extend_values(value_t *valuep, int num)
{
  value_t value;
  int r, n = 0;

  for (; *valuep; valuep = &(*valuep)->next)
    n++;

  while (n < num) {
    if ((r = value_alloc(&value)) < 0)
      return r;
    value->next = *valuep;
    *valuep = value;
    n++;
  }

  return 0;
}

int value_copy_value(value_t dvalue, value_t svalue)
{
  int r;
  value_t value;

  value = value_entity(svalue);

  switch (value->type) {
  case VALUE_TYPE_INTEGER:
    if ((r = value_set_integer(dvalue, value->val.integer)) < 0)
      return r;
    break;
  case VALUE_TYPE_STRING:
    if ((r = value_set_string(dvalue, value->val.string->s, value->val.string->length)) < 0)
      return r;
    break;
#ifdef NLL_FLOATING_POINT
  case VALUE_TYPE_FLOAT:
    if ((r = value_set_float(dvalue, value->val.floating)) < 0)
      return r;
    break;
#endif
  case VALUE_TYPE_AREA:
    if ((r = value_set_area(dvalue, value->val.area)) < 0)
      return r;
    break;
  case VALUE_TYPE_POINTER:
    if ((r = value_set_pointer(dvalue, value->val.pointer.point, value->val.pointer.offset)) < 0)
      return r;
    break;
  case VALUE_TYPE_UNDEFINED:
    break;
  case VALUE_TYPE_ARRAY:
  default:
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
  }

  return 0;
}

int value_copy_values(value_t dvalue, value_t svalue)
{
  int r;

  for (; dvalue && svalue; dvalue = dvalue->next, svalue = svalue->next) {
    if ((r = value_copy_value(dvalue, svalue)) < 0)
      return r;
  }

  return 0;
}

value_type_t value_get_type(value_t value)
{
  value = value_entity(value);
  return value->type;
}

unsigned int value_get_flags(value_t value)
{
  value = value_entity(value);
  return value->flags;
}

int value_set_flags(value_t value, unsigned int flags)
{
  value = value_entity(value);
  value->flags |= flags;
  return 0;
}

int value_reset_flags(value_t value, unsigned int flags)
{
  value = value_entity(value);
  value->flags &= ~flags;
  return 0;
}

int value_is_same(value_t value0, value_t value1)
{
  value0 = value_entity(value0);
  value1 = value_entity(value1);
  return (value0 == value1) ? 1 : 0;
}

int value_get_value(value_t value, value_t *valuep)
{
  if (value->type != VALUE_TYPE_VALUE)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *valuep = value->val.value;

  return 0;
}

int value_set_value(value_t value, value_t _value)
{
  int r;

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_VALUE)) {
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
  }

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->flags & VALUE_FLAG_CONST))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  _value->refcount++;

  if (value->type == VALUE_TYPE_VALUE) {
    if ((r = value_free(value->val.value)) < 0)
      return r;
  }

  value->type = VALUE_TYPE_VALUE;
  value->val.value = _value;

  return 0;
}

int value_set_values(value_t value, value_t _value)
{
  int r;

  for (; value && _value; value = value->next, _value = _value->next) {
    if ((r = value_set_value(value, _value)) < 0)
      return r;
  }

  return 0;
}

int value_get_array(value_t value, int *numberp, array_t *arrayp)
{
  value = value_entity(value);

  if (value->type != VALUE_TYPE_ARRAY)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *numberp = value->val.array.number;
  *arrayp = value->val.array.array;

  return 0;
}

int value_set_array(value_t value, int number, array_t array)
{
  value = value_entity(value);

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_ARRAY)) {
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
  }

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->flags & VALUE_FLAG_CONST))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  value->type = VALUE_TYPE_ARRAY;
  value->val.array.number = number;
  value->val.array.array = array;

  return 0;
}

int value_get_integer(value_t value, integer_t *integerp)
{
  value = value_entity(value);

  if (value->type != VALUE_TYPE_INTEGER)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *integerp = value->val.integer;

  return 0;
}

int value_set_integer(value_t value, integer_t integer)
{
#ifdef NLL_TYPECHANGE_ENABLE
  int r;
#endif

  value = value_entity(value);

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_INTEGER)) {
    if (value->flags & VALUE_FLAG_STATIC)
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
#ifdef NLL_TYPECHANGE_ENABLE
    if ((r = value_clear(value)) < 0)
      return r;
#else
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
  }

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->flags & VALUE_FLAG_CONST))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  value->type = VALUE_TYPE_INTEGER;
  value->val.integer = integer;

  return 0;
}

int value_get_string(value_t value, char **stringp, int *lengthp)
{
  value = value_entity(value);

  if (value->type != VALUE_TYPE_STRING)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *stringp = value->val.string->s;
  if (lengthp)
    *lengthp = value->val.string->length;

  return 0;
}

int value_set_string(value_t value, char *string, int length)
{
  int r;

  value = value_entity(value);

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_STRING)) {
    if (value->flags & VALUE_FLAG_STATIC)
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
#ifdef NLL_TYPECHANGE_ENABLE
    if ((r = value_clear(value)) < 0)
      return r;
#else
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
  }

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->flags & VALUE_FLAG_CONST))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  if (length > STRING_MAXLEN)
    return NLL_ERRCODE_STRING_TOO_LONG;

  if (value->type == VALUE_TYPE_UNDEFINED) {
    if ((r = string_alloc(&value->val.string)) < 0)
      return r;
  }

  value->type = VALUE_TYPE_STRING;
  if ((r = string_set(value->val.string, string, length)) < 0)
    return 0;

  return 0;
}

#ifdef NLL_FLOATING_POINT
int value_get_float(value_t value, double *floatingp)
{
  value = value_entity(value);

  if (value->type != VALUE_TYPE_FLOAT)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *floatingp = value->val.floating;

  return 0;
}

int value_set_float(value_t value, double floating)
{
#ifdef NLL_TYPECHANGE_ENABLE
  int r;
#endif

  value = value_entity(value);

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_FLOAT)) {
    if (value->flags & VALUE_FLAG_STATIC)
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
#ifdef NLL_TYPECHANGE_ENABLE
    if ((r = value_clear(value)) < 0)
      return r;
#else
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
  }

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->flags & VALUE_FLAG_CONST))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  value->type = VALUE_TYPE_FLOAT;
  value->val.floating = floating;

  return 0;
}
#endif

int value_get_area(value_t value, area_t *areap)
{
  value = value_entity(value);

  if (value->type != VALUE_TYPE_AREA)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *areap = value->val.area;

  return 0;
}

int value_set_area(value_t value, area_t area)
{
  int r;

  value = value_entity(value);

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_AREA)) {
    if (value->flags & VALUE_FLAG_STATIC)
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
#ifdef NLL_TYPECHANGE_ENABLE
    if ((r = value_clear(value)) < 0)
      return r;
#else
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
  }

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->flags & VALUE_FLAG_CONST))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  if (area)
    area->refcount++;

  if (value->type == VALUE_TYPE_AREA) {
    if ((r = area_free(value->val.area)) < 0)
      return r;
  }

  value->type = VALUE_TYPE_AREA;
  value->val.area = area;

  return 0;
}

int value_get_pointer(value_t value, value_t *pointp, int *offsetp)
{
  value = value_entity(value);

  if (value->type != VALUE_TYPE_POINTER)
    return NLL_ERRCODE_VALUE_INVALID_TYPE;

  *pointp = value->val.pointer.point;
  if (offsetp)
    *offsetp = value->val.pointer.offset;

  return 0;
}

int value_set_pointer(value_t value, value_t point, int offset)
{
  int r;

  value = value_entity(value);

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->type != VALUE_TYPE_POINTER)) {
    if (value->flags & VALUE_FLAG_STATIC)
      return NLL_ERRCODE_VALUE_INVALID_TYPE;
#ifdef NLL_TYPECHANGE_ENABLE
    if ((r = value_clear(value)) < 0)
      return r;
#else
    return NLL_ERRCODE_VALUE_INVALID_TYPE;
#endif
  }

  if ((value->type != VALUE_TYPE_UNDEFINED) &&
      (value->flags & VALUE_FLAG_CONST))
    return NLL_ERRCODE_VALUE_CONST_VALUE;

  if (point)
    point->refcount++;

  if (value->type == VALUE_TYPE_POINTER) {
    if (value->val.pointer.point) {
      if ((r = value_free(value->val.pointer.point)) < 0)
	return r;
    }
  }

  value->type = VALUE_TYPE_POINTER;
  value->val.pointer.point = point;
  value->val.pointer.offset = point ? offset : 0;

  return 0;
}

int value_get_integer_float(value_t value, integer_t *integerp)
{
  int r;
#ifdef NLL_FLOATING_POINT
  double f;
#endif

  if ((r = value_get_integer(value, integerp)) < 0) {
#ifdef NLL_FLOATING_POINT
    if ((r = value_get_float(value, &f)) < 0)
      return r;
    *integerp = f;
#else
    return r;
#endif
  }

  return 0;
}

#ifdef NLL_FLOATING_POINT
int value_get_float_integer(value_t value, double *floatingp)
{
  int r;
  integer_t integer;

  if ((r = value_get_float(value, floatingp)) < 0) {
    if ((r = value_get_integer(value, &integer)) < 0)
      return r;
    *floatingp = integer;
  }

  return 0;
}
#endif

int value_output_value(value_t value)
{
  char buf[32];

  value = value_entity(value);

  switch (value->type) {
  case VALUE_TYPE_INTEGER:
    sprintf(buf, "%ld", value->val.integer);
    framebuf_puts(buf);
    break;
  case VALUE_TYPE_STRING:
    framebuf_puts(value->val.string->s);
    break;
#ifdef NLL_FLOATING_POINT
  case VALUE_TYPE_FLOAT:
    sprintf(buf, "%f", value->val.floating);
    framebuf_puts(buf);
    break;
#endif
  case VALUE_TYPE_ARRAY:
  case VALUE_TYPE_AREA:
  case VALUE_TYPE_POINTER:
  case VALUE_TYPE_UNDEFINED:
  default:
    break;
  }

  return 0;
}

int value_output_values(value_t value, const char *separator)
{
  int printed = 0;

  if (!separator) separator = "";

  for (; value; value = value->next) {
    switch (value_get_type(value)) {
    case VALUE_TYPE_INTEGER:
    case VALUE_TYPE_STRING:
#ifdef NLL_FLOATING_POINT
    case VALUE_TYPE_FLOAT:
#endif
      if (printed)
	framebuf_puts(separator);
      else
	printed = 1;
      break;
    default:
      break;
    }
    value_output_value(value);
  }

  return 0;
}

int value_dump(FILE *fp, value_t value)
{
  value_t v = NULL;
  int i;

  for (v = value; value; value = value->next) {
    nll_wait_output(fp);
    if (v != value)
      fprintf(fp, ",");
    if (value->flags & VALUE_FLAG_CONST) {
      switch (value->type) {
      case VALUE_TYPE_INTEGER:
      case VALUE_TYPE_STRING:
#ifdef NLL_FLOATING_POINT
      case VALUE_TYPE_FLOAT:
#endif
	break;
      default:
	fprintf(fp, "(C)");
      }
    }
    if (value->flags & VALUE_FLAG_STATIC)
      fprintf(fp, "(S)");
    switch (value->type) {
    case VALUE_TYPE_VALUE:
      fprintf(fp, "(");
      value_dump(fp, value->val.value);
      fprintf(fp, ")");
      break;
    case VALUE_TYPE_ARRAY:
      fprintf(fp, "ARRAY[%d](", value->val.array.number);
      for (i = 0; i < value->val.array.number; i++) {
	fprintf(fp, "[%d]", i);
	value_dump(fp, value->val.array.array[i].value);
      }
      fprintf(fp, ")");
      break;
    case VALUE_TYPE_INTEGER:
      fprintf(fp, "%ld", value->val.integer);
      break;
    case VALUE_TYPE_STRING:
      string_print(fp, value->val.string);
      break;
#ifdef NLL_FLOATING_POINT
    case VALUE_TYPE_FLOAT:
      fprintf(fp, "%f(FLOAT)", value->val.floating);
      break;
#endif
    case VALUE_TYPE_AREA:
      fprintf(fp, "AREA(%d)", value->val.area->size);
      area_print(fp, value->val.area);
      break;
    case VALUE_TYPE_POINTER:
      fprintf(fp, "&");
      if (value->val.pointer.point) {
	value_dump(fp, value->val.pointer.point);
	fprintf(fp, "[%d]", value->val.pointer.offset);
      } else {
	fprintf(fp, "NULL");
      }
      break;
    case VALUE_TYPE_UNDEFINED:
    default:
      fprintf(fp, "UNDEF");
      break;
    }
  }

  fflush(fp);

  return 0;
}
